###############################################################################################################
########################################## Basic resolve functions ############################################
###############################################################################################################

"""
	resolve!(t::Tree, S::SplitList; conflict=:fail, usemask=false, tau=0.)

Add splits in `S` to `t` by introducing internal nodes.
New nodes are assigned a time `tau` (`0` by default).
If `conflict != :ignore`, will fail if a split `s` in `S` is not compatible with `t`.
Otherwise, silently skip the conflicting splits.
"""
function resolve!(
	t::Tree{T}, S::SplitList;
	conflict=:fail, usemask=false, tau=0., safe=false
) where T
	# Label for created nodes
	label_i = parse(Int64, TreeTools.create_label(t, "RESOLVED")[10:end])
	#
	tsplits = SplitList(t)
	for (i,s) in enumerate(S)
		if !safe && !in(s, tsplits; usemask)
			if iscompatible(s, tsplits; usemask)
				if usemask
					roots = TreeTools.blca([t.lleaves[x] for x in leaves(S,i)]...)
				else
					roots = TreeTools.blca([t.lleaves[x] for x in leaves(S,i)]...)
				end
				R = lca(roots)
				# Creating a new node with `roots` as children and `r` as ancestor.
				nr = TreeNode(T(), label="RESOLVED_$(label_i)")
				label_i += 1
				for r in roots
					prunenode!(r)
					graftnode!(nr,r)
				end
				graftnode!(R, nr, tau=tau)
				push!(tsplits.splits, s)
			elseif conflict != :ignore
				error("Tried to resolve tree with an incompatible split.")
			end
		end
	end
	node2tree!(t, t.root)
end

"""
	resolve(trees::Dict{T, <:Tree}, splits::Dict{T, <:SplitList}; kwargs...) where T

Resolve `trees[s]` with splits in `splits[s]` by calling `resolve!`.
Return dictionary of resolved trees. `trees` and `splits` must share keys.
This is meant to be used for dictionaries of trees/splits indexed by flu segments.
"""
function resolve(trees::Dict{T, <:Tree}, splits::Dict{T, <:SplitList}; kwargs...) where T
	resolved_trees = Dict(k => copy(t) for (k,t) in trees)
	for (s,S) in splits
		resolve!(resolved_trees[s], S; kwargs...)
	end
	return resolved_trees
end


"""
	resolve!(S1::SplitList, t1::Tree, S2::SplitList)

Add splits of `S2` in `S1` if they resolve `t1`.
"""
function resolve!(S1new, S1::SplitList, t1::Tree, S2::SplitList)
	c = 0
	for s2 in S2
		r1 = t1.lleaves[S2.leaves[s2.dat[1]]]
		for i in 2:length(s2.dat)
			r1 = lca(r1, t1.lleaves[S2.leaves[s2.dat[i]]])
		end
		#r1 = lca(t1, S2.leaves[s2.dat]) # Ancestor of nodes in s2 in t1
		s1 = S1.splitmap[r1.label]
		if s1 != s2 && !in(s2, S1new) && arecompatible(s1, s2)
			# Consider the set of splits just below r1 that are subsplits of s2
			# If I join those, I should get exactly s2
			# Otherwise, can't use s2 to resolve r1
			stmp = Split(0)
			for n in r1.child
				if n.isleaf
					i = findfirst(==(n.label), S2.leaves)
					if in(i, s2.dat)
						TreeTools.joinsplits!(stmp, Split([i]))
					end
				else
					if TreeTools.is_sub_split(S1.splitmap[n.label], s2)
						TreeTools.joinsplits!(stmp, S1.splitmap[n.label])
					end
				end
			end

			if stmp == s2
				push!(S1.splits, s2)
				push!(S1new.splits, s2)
				c += 1
			end
		end
	end

	return c
end

function resolve!(S1::SplitList, S2::SplitList, t1::Tree, t2::Tree)
	nit = 0
	nitmax = 20
	flag = true
	S1new = SplitList(S1.leaves)
	S2new = SplitList(S2.leaves)
	while flag && nit < nitmax
		flag = false
		c = resolve!(S1new, S1, t1, S2)
		c != 0 && (flag = true)
		c = resolve!(S2new, S2, t2, S1)
		c != 0 && (flag = true)
		nit += 1
	end

	if nit == nitmax
		@warn "Maximum number of iterations reached"
	end

	return [S1new, S2new]
end

###############################################################################################################
################################### Resolve trees using eachother splits ######################################
###############################################################################################################

"""
	resolve!(t1::Tree, t2::Tree; tau=0.)

Resolve `t1` using splits of `t2` and inversely.
Every split of `t2` a tree that is compatible with `t1` is introduced in `t1` with branch
length `tau` (and inversely). Return new splits in each tree.
"""
function resolve!(t1::Tree, t2::Tree; tau=0.)
	S = [SplitList(t) for t in (t1,t2)]
	Snew = resolve!(S[1], S[2], t1, t2)
	for (t, s) in zip((t1,t2), S)
		resolve!(t, s, conflict=:fail, usemask=false, tau=tau)
	end

	return Snew
end




###############################################################################################################
############################## Resolve trees using inferred compatible clades #################################
###############################################################################################################

"""
	resolve!(t1::Tree{MiscData}, t2::Tree{MiscData}, MCCs; tau = 0.)

Resolve `t1` using `t2` and inversely using the list of MCCs.
New branches have a length `tau`.
Return the list of resolved splits in each tree.
"""

function resolve!(t1::Tree{T}, t2::Tree{T}, MCCs; tau = 0., strict=false) where T 
	resolvable_splits = TreeKnit.new_splits(MCCs, t1, t2; strict)
	resolve!(t1, resolvable_splits[1]; conflict=:fail, tau)
	resolve!(t2, resolvable_splits[2]; conflict=:fail, tau)

	return resolvable_splits
end


